/* This file is part of the "Regression Tree Fields" (RTF) source code distribution,
 * obtained from http://www.gris.tu-darmstadt.de/research/visinf/software.
 * It is provided to you under the terms of the Microsoft Research License Agreement
 * (MSR-LA). Please see License.txt for details.
 *
 *
 * File: Dataset.h
 * Implements the dataset class of the image deconvolution example.
 *
 * ---
 * Last modified by Uwe Schmidt on 07/05/2014.
 */

#ifndef H_DECONVOLUTION_DATASET_H
#define H_DECONVOLUTION_DATASET_H

#include <string>
#include <random>
#include <vector>
#include <cassert>
#include <fstream>
#include <iostream>
#include <exception>
#include <algorithm>

#include <boost/filesystem.hpp>

#include <itkImage.h>
#include <itkZeroFluxNeumannPadImageFilter.h>
#include <itkCropImageFilter.h>

#include "RTF/Image.h"
#include "RTF/Array.h"
#include "RTF/Unary.h"
#include "RTF/Types.h"
#include "RTF/Training.h"

 // Use the filterbank by Qi Gao and Stefan Roth, DAGM 2012.
#include "QiDAGM12Filterbank.h"
#define Filterbank QiDAGM12Filterbank

namespace Deconvolution
{
    class IOException : public std::exception
    {
    private:
        std::string w;
    public:
        IOException(const std::string& what_) : w(what_) {}

        virtual const char *what() const throw()
        {
            return w.c_str();
        }

        virtual ~IOException() throw() {}
    };

    // The dataset class for deconvolution; methods and typedefs marked by INTERFACE
    // are required by the RTF code and must be implemented by any dataset class.
    class Dataset
    {
    public:

        // INTERFACE: define the output label of each pixel
        typedef Training::LabelVector<double, 1>   UnaryGroundLabel;
        typedef Training::LabelVector<double, 2>   PairwiseGroundLabel;

        // INTERFACE: define the type of data stored for each input pixel
        typedef Training::LabelVector<float,
                1 + // Original grayscale input
#ifdef REPRODUCE_CVPR13_RESULTS
                5 + // Deblurred output of up to 5 previous layers
#else
                1 + // Deblurred output of previous layer
#endif
                1 * Filterbank::filter_count + // Filter responses of input or output from directly preceding layer
                1   // Blur kernel
                > InputLabel;


        static size_t filter_input_index(size_t level)
        {
#ifdef REPRODUCE_CVPR13_RESULTS
            assert(level >= 0 && level <= 5);
            return level;
#else
            // if this is the first layer in the stack, we simply use the original input image
            // otherwise, we use the prediction of the previous layer
            return level == 0 ? 0 : 1;
#endif
        }

        static size_t filter_output_index(size_t fi)
        {
#ifdef REPRODUCE_CVPR13_RESULTS
            return 6 + fi;
#else
            return 2 + fi;
#endif
        }

        static size_t blur_kernel_index()
        {
            return InputLabel::Size-1;
        }

        Dataset(const std::string& path_, const std::string& type_)
            : path(path_), type(type_)
        {
            ReadDescriptor();
        }

        void ReleaseCache() const
        {
            for( size_t i = 0; i < GetImageCount(); ++i )
            {
                groundTruthImages[i] = ImageRefC<UnaryGroundLabel>();
                inputImages[i] = ImageRef<InputLabel>();
                kernels[i] = ImageRefC<UnaryGroundLabel>();
            }
        }

        // INTERFACE: return the number of images in the dataset
        size_t GetImageCount() const
        {
            return inputImages.size();
        }

        std::string GetImageName(size_t idx) const
        {
            assert( idx < fileNames.size() );
            return fileNames[idx];
        }

        ImageRefC<UnaryGroundLabel> GetKernel(size_t idx) const
        {
            assert(idx < kernels.size());
            if( ! kernels[idx] )
            {
                kernels[idx] = LoadKernelDLM(GetImagePath(idx, "kernels"));
                // kernel size cannot be even
                assert((kernels[idx].Height()%2==1) &&
                       (kernels[idx].Width() %2==1));
            }
            return kernels[idx];
        }

        // INTERFACE: return the idx'th ground truth image
        ImageRefC<UnaryGroundLabel> GetGroundTruthImage(size_t idx) const
        {
            assert(idx < groundTruthImages.size());
            if( ! groundTruthImages[idx] )
            {
                groundTruthImages[idx] = LoadGroundTruthImage(GetImagePath(idx, "labels"), GetKernel(idx));
            }
            return groundTruthImages[idx];
        }

        void SavePrediction(const VecCRef<ImageRefC<UnaryGroundLabel>>& prediction, size_t idx, const std::string& path) const
        {
            auto cropped = CropImage(prediction[idx], GetKernel(idx));
            if( path.back() == 'g' || path.back() == 'G' )
                SaveGroundTruthImagePNG(cropped, path);
            else
                SaveGroundTruthImageDLM(cropped, path);
        }        

        static void SaveGroundTruthImage(const ImageRefC<UnaryGroundLabel>& ground, const std::string& path)
        {
            if( path.back() == 'g' || path.back() == 'G' )
                SaveGroundTruthImagePNG(ground, path);
            else
                SaveGroundTruthImageDLM(ground, path);
        }

        // INTERFACE: return the idx'th input image.
        ImageRefC<InputLabel> GetInputImage(size_t idx) const
        {
            assert(idx < inputImages.size());
            if( ! inputImages[idx] )
            {
                inputImages[idx] = LoadInputImage(GetImagePath(idx, "images"), GetKernel(idx));
            }
            return inputImages[idx];
        }

        static void SaveInputImage(const ImageRefC<InputLabel>& input, const std::string& path)
        {
            if( path.back() == 'g' || path.back() == 'G' )
                return SaveInputImagePNG(input, path);
            else
                return SaveInputImageDLM(input, path);
        }

        static void SaveGroundTruthImagePNG(const ImageRefC<UnaryGroundLabel>& ground, const std::string& path)
        {
            ImageRef<unsigned char, 1> img(ground.Width(), ground.Height());
            for( int y = 0; y < img.Height(); ++y )
                for( int x = 0; x < img.Width(); ++x )
                    for( int c = 0; c < 1; ++c )
                        *(img.Ptr(x,y) + c) = (unsigned char) (std::max(0.0f, std::min((float) ground(x,y)[c], 1.0f))*255.0f + .5f );
            Utility::WritePNG(img, path);
        }

        static void SaveGroundTruthImageDLM(const ImageRefC<UnaryGroundLabel>& ground, const std::string& path)
        {
            ImageRef<float, 1> img(ground.Width(), ground.Height());
            for( int y = 0; y < img.Height(); ++y )
                for( int x = 0; x < img.Width(); ++x )
                    for( int c = 0; c < 1; ++ c )
                        *(img.Ptr(x,y)+c) = (float) ground(x,y)[c];
            Utility::WriteDLM(img, path);
        }

        static void SaveInputImagePNG(const ImageRefC<InputLabel>& input, const std::string& path)
        {
            ImageRef<unsigned char, 1> img(input.Width(), input.Height());
            for( int y = 0; y < img.Height(); ++y )
                for( int x = 0; x < img.Width(); ++x )
                    for( int c = 0; c < 1; ++ c )
                        *(img.Ptr(x,y)+c) = (unsigned char) ( std::max(0.0f, std::min((float) input(x,y)[c], 1.0f))*255.0f + .5f );
            Utility::WritePNG(img, path);
        }

        static void SaveInputImageDLM(const ImageRefC<InputLabel>& input, const std::string& path)
        {
            ImageRef<float, 1> img(input.Width(), input.Height());
            for( int y = 0; y < img.Height(); ++y )
                for( int x = 0; x < img.Width(); ++x )
                    for( int c = 0; c < 1; ++ c )
                        *(img.Ptr(x,y)+c) = input(x,y)[c];
            Utility::WriteDLM(img, path);
        }

        std::string GetImagePath(size_t idx, const std::string& prefix) const
        {
            assert(idx < GetImageCount());

            boost::filesystem::path dlmPath = path;
            dlmPath = dlmPath / prefix / (fileNames[idx] + ".dlm");
            boost::filesystem::path pngPath = path;
            pngPath = pngPath / prefix / (fileNames[idx] + ".png");

            if( boost::filesystem::exists(dlmPath) )
            {
                return dlmPath.string();
            }
            else
            {
                if( ! boost::filesystem::exists(pngPath) )
                    throw IOException("Can't find " + prefix + " image: " + pngPath.string());
                else
                    return pngPath.string();
            }
            return pngPath.string();
        }

        static ImageRefC<UnaryGroundLabel> LoadGroundTruthImage(const std::string& path, const ImageRefC<UnaryGroundLabel> kernel)
        {
            // Check if we load a PNG or a DLM file
            if( path.back() == 'g' || path.back() == 'G' )
                return LoadGroundTruthImagePNG(path, kernel);
            else
                return LoadGroundTruthImageDLM(path, kernel);
        }

        static ImageRefC<UnaryGroundLabel> LoadGroundTruthImagePNG(const std::string& path, const ImageRefC<UnaryGroundLabel> kernel)
        {
            auto img = Utility::ReadPNG<unsigned char, 1>(path);
            // pad image
            img = PadImageReplicate(img, kernel);
            ImageRef<UnaryGroundLabel> ret(img.Width(), img.Height());

            for( int y = 0; y < ret.Height(); ++y )
                for( int x = 0; x < ret.Width(); ++x )
                    for( int c = 0; c < 1; ++c )
                        ret(x,y)[c] = *(img.Ptr(x,y)+c) / 255.0;
            return ret;
        }

        static ImageRefC<UnaryGroundLabel> LoadKernelDLM(const std::string& path)
        {
            auto img = Utility::ReadDLM<float, 1>(path);
            ImageRef<UnaryGroundLabel> ret(img.Width(), img.Height());
            for( int y = 0; y < ret.Height(); ++y )
                for( int x = 0; x < ret.Width(); ++x )
                    for( int c = 0; c < 1; ++c )
                        ret(x,y)[c] = *(img.Ptr(x,y)+c);
            return ret;
        }

        static ImageRefC<UnaryGroundLabel> LoadGroundTruthImageDLM(const std::string& path, const ImageRefC<UnaryGroundLabel> kernel)
        {
            auto img = Utility::ReadDLM<float, 1>(path);
            // pad image
            img = PadImageReplicate(img, kernel);
            ImageRef<UnaryGroundLabel> ret(img.Width(), img.Height());

            for( int y = 0; y < ret.Height(); ++y )
                for( int x = 0; x < ret.Width(); ++x )
                    for( int c = 0; c < 1; ++c )
                        ret(x,y)[c] = *(img.Ptr(x,y)+c);
            return ret;
        }

        ImageRef<InputLabel> LoadInputImage(const std::string& path, const ImageRefC<UnaryGroundLabel> kernel) const
        {
            if( path.back() == 'g' || path.back() == 'G' )
                return LoadInputImagePNG(path, kernel);
            else
                return LoadInputImageDLM(path, kernel);
        }

        template<typename TValue>
        static ImageRef<TValue,1> PadImageReplicate(const ImageRef<TValue,1> img, const ImageRefC<UnaryGroundLabel> kernel)
        {
            const auto cx = img.Width(), cy = img.Height();
            const auto ckx = kernel.Width(), cky = kernel.Height();
            const auto bx = (ckx-1)/2, by = (cky-1)/2;

            typedef itk::Image<float, 2> ImageType;
            ImageType::Pointer image = ImageType::New();

            ImageType::IndexType start = {{0,0}};
            ImageType::SizeType size;
            size[0] = cx;
            size[1] = cy;

            ImageType::RegionType region;
            region.SetSize(size);
            region.SetIndex(start);

            image->SetRegions(region);
            image->Allocate();
            
            // copy data
            for(int y = 0; y < cy; y++) {
              for(int x = 0; x < cx; x++) {
                ImageType::IndexType idx = {{x,y}};
                image->SetPixel(idx, img(x,y));
              }
            }

            typedef itk::ZeroFluxNeumannPadImageFilter<ImageType,ImageType> FilterType;
            FilterType::Pointer pad_filter = FilterType::New();
            pad_filter->SetInput(image);

            ImageType::SizeType pad_size;
            pad_size[0] = bx;
            pad_size[1] = by;
            pad_filter->SetPadBound(pad_size);

            pad_filter->Update();
            auto padded = pad_filter->GetOutput();

            // subtract padded boundary extent from pixel indices
            // because PadImageFilter changes the Index of the image
            
            const auto padded_size = padded->GetLargestPossibleRegion().GetSize();
            ImageRef<TValue,1> ret(padded_size[0], padded_size[1]);
            
            for( int y = 0; y < ret.Height(); ++y ) {
                for( int x = 0; x < ret.Width(); ++x ) {
                    ImageType::IndexType idx = {{x-bx,y-by}};
                    ret(x,y) = padded->GetPixel(idx);
                }
            }
            return ret;
        }

        static ImageRefC<UnaryGroundLabel> CropImage(const ImageRefC<UnaryGroundLabel> img, const ImageRefC<UnaryGroundLabel> kernel)
        {
            const auto cx = img.Width(), cy = img.Height();
            const auto ckx = kernel.Width(), cky = kernel.Height();
            const auto bx = (ckx-1)/2, by = (cky-1)/2;

            typedef itk::Image<float, 2> ImageType;
            ImageType::Pointer image = ImageType::New();

            ImageType::IndexType start = {{0,0}};
            ImageType::SizeType size;
            size[0] = cx;
            size[1] = cy;

            ImageType::RegionType region;
            region.SetSize(size);
            region.SetIndex(start);

            image->SetRegions(region);
            image->Allocate();
            
            // copy data
            for(int y = 0; y < cy; y++) {
              for(int x = 0; x < cx; x++) {
                ImageType::IndexType idx = {{x,y}};
                image->SetPixel(idx, img(x,y)[0]);
              }
            }

            typedef itk::CropImageFilter <ImageType,ImageType> FilterType;
            FilterType::Pointer crop_filter = FilterType::New();
            crop_filter->SetInput(image);

            ImageType::SizeType crop_size;
            crop_size[0] = bx;
            crop_size[1] = by;
            crop_filter->SetBoundaryCropSize(crop_size);

            crop_filter->Update();
            auto cropped = crop_filter->GetOutput();

            // add padded boundary extent to pixel indices
            // because CropImageFilter changes the Index of the image

            const auto cropped_size = cropped->GetLargestPossibleRegion().GetSize();
            ImageRef<UnaryGroundLabel> ret(cropped_size[0], cropped_size[1]);

            for( int y = 0; y < ret.Height(); ++y ) {
                for( int x = 0; x < ret.Width(); ++x ) {
                    ImageType::IndexType idx = {{x+bx,y+by}};
                    ret(x,y)[0] = cropped->GetPixel(idx);
                }
            }
            return ret;
        }        

        ImageRef<InputLabel> LoadInputImagePNG(const std::string& path, const ImageRefC<UnaryGroundLabel> kernel) const
        {
            auto img = Utility::ReadPNG<unsigned char, 1>(path);
            // pad image
            img = PadImageReplicate(img, kernel);
            ImageRef<InputLabel> ret(img.Width(), img.Height());

            for( int y = 0; y < ret.Height(); ++y )
                for( int x = 0; x < ret.Width(); ++x )
                    for( int c = 0; c < 1; ++c )
                        ret(x,y)[c] = *(img.Ptr(x,y)+c) / 255.0f;
            return PreProcessInput(0,ret,kernel);
        }

        ImageRef<InputLabel> LoadInputImageDLM(const std::string& path, const ImageRefC<UnaryGroundLabel> kernel) const
        {
            auto img = Utility::ReadDLM<float, 1>(path);
            // pad image
            img = PadImageReplicate(img, kernel);
            ImageRef<InputLabel> ret(img.Width(), img.Height());

            for( int y = 0; y < ret.Height(); ++y )
                for( int x = 0; x < ret.Width(); ++x )
                    for( int c = 0; c < 1; ++c )
                        ret(x,y)[c] = *(img.Ptr(x,y)+c);
            return PreProcessInput(0,ret,kernel);
        }

        // INTERFACE: in a stacked RTF ensemble, perform pre-processing before each new layer
        void InitializeForCascadeLevel(size_t level, VecCRef<ImageRefC<UnaryGroundLabel>> previousPrediction) const
        {
            if( level > 0 ) // Initial pre-processing is done already while first loading the input images
                for( size_t idx = 0; idx < GetImageCount(); ++idx )
                    PreProcessInput(level, inputImages[idx], kernels[idx], previousPrediction[idx]);
        }

        static ImageRef<InputLabel>
        PreProcessInput(size_t level,
                        ImageRef<InputLabel> input,
                        ImageRefC<UnaryGroundLabel> kernel,
                        ImageRefC<UnaryGroundLabel> previousPrediction = ImageRef<UnaryGroundLabel>(0,0))
        {
            const int cx = input.Width(), cy = input.Height();

            // has to be done only once
            if (level == 0) {
                // 1) copy the blur kernel into InputLabel
                const int ckx = kernel.Width(), cky = kernel.Height();
                // kernel must be smaller than image
                assert(cx > ckx && cy > cky);

                // size of blur kernel in last two pixels
                input(cx-2,cy-1)[blur_kernel_index()] = static_cast<double>(ckx);
                input(cx-1,cy-1)[blur_kernel_index()] = static_cast<double>(cky);

                // copy kernel
                #pragma omp parallel for
                for( int y = 0; y < cky; ++y )
                    for( int x = 0; x < ckx; ++x )
                        input(x,y)[blur_kernel_index()] = kernel(x,y)[0];
            }

            // 2) copy previousPrediction to its correct place
            if (level > 0) {
                assert(previousPrediction.Width() > 0);

                #pragma omp parallel for
                for( int y = 0; y < cy; ++y )
                    for( int x = 0; x < cx; ++x )
                        for( int c = 0; c < 1; ++c )
                            input(x,y)[filter_input_index(level)] = previousPrediction(x,y)[c];
            }

            return UpdateFilterResponses(level, input);
        }

        static ImageRef<InputLabel>
        UpdateFilterResponses(size_t level, ImageRef<InputLabel> input)
        {
            // Computes a 2D convolution with each filter
            // and stores the response in the input image for later use in the linear basis and the feature checks
            const int cx = input.Width();
            const int cy = input.Height();

            const int fcy_offset = Filterbank::filter_size_y/2;
            const int fcx_offset = Filterbank::filter_size_x/2;

            std::vector<float> filter_values(Filterbank::filter_size_y * Filterbank::filter_size_x * Filterbank::filter_count);
            int idx = 0;
            for (int fy = 0; fy < Filterbank::filter_size_y; ++fy)
                for (int fx = 0; fx < Filterbank::filter_size_x; ++fx)
                    for (int fi = 0; fi < Filterbank::filter_count; ++fi)
                        filter_values[idx++] = (float) Filterbank::filter_values[fi][fy][fx];

            #pragma omp parallel for
            for (int y = 0; y < cy; ++y)
            {
                for (int x = 0; x < cx; ++x)
                {
                    const float* filter_ptr = &(filter_values[0]);

                    auto &output_label = input(x,y);
                    memset(&output_label[filter_output_index(0)], 0, 1 * Filterbank::filter_count * sizeof(float));

                    for (int fy = 0; fy < Filterbank::filter_size_y; ++fy)
                    {
                        int eff_y = y+fy-fcy_offset;
                        eff_y = std::max(0, std::min(eff_y, cy-1));

                        for (int fx = 0; fx < Filterbank::filter_size_x; ++fx)
                        {
                            int eff_x = x+fx-fcx_offset;
                            eff_x = std::max(0, std::min(eff_x, cx-1));

                            const float* const input_ptr = &(input(eff_x, eff_y)[filter_input_index(level)]);
                            float* output_ptr = &output_label[filter_output_index(0)];

                            for (int fi = 0; fi < Filterbank::filter_count; ++fi)
                            {
                                const auto filter_value = *filter_ptr++;

                                for( int c = 0; c < 1; ++c )
                                    *output_ptr++ += input_ptr[c] * filter_value;
                            }
                        }
                    }
                }
            }

            return input;
        }


    private:

#ifdef USE_MPI
        void ReadDescriptor()
        {
            boost::filesystem::path p = path;
            const std::string dpath = (p / (type + ".txt")).string();
            std::ifstream ifs(dpath);

            std::cerr << "reading " << dpath << std::endl;

            if( ifs.fail() )
                throw IOException("failed to open '" + dpath + "'");

            std::string file;
            size_t line = 0;
            while(std::getline(ifs, file))
            {
                if( (file != "") && (line++ % MPI::Communicator().size() == MPI::Communicator().rank()) )
                {
                    fileNames.push_back(file);
                }
            }
            inputImages.resize(fileNames.size());
            groundTruthImages.resize(fileNames.size());
            kernels.resize(fileNames.size());
        }
#else
        void ReadDescriptor()
        {
            boost::filesystem::path p = path;
            const std::string dpath = (p / (type + ".txt")).string();

            std::ifstream ifs(dpath);

            if( ifs.fail() )
                throw IOException("failed to open '" + dpath + "'");

            std::string file;
            while(std::getline(ifs, file))
            {
                if( file != "" )
                    fileNames.push_back(file);
            }
            groundTruthImages.resize(fileNames.size());
            inputImages.resize(fileNames.size());
            kernels.resize(fileNames.size());
        }
#endif

        const std::string                                path;
        const std::string                                type;

        std::vector<std::string>                         fileNames;
        mutable std::vector<ImageRefC<UnaryGroundLabel>> groundTruthImages;
        mutable std::vector<ImageRef<InputLabel>>        inputImages;
        mutable std::vector<ImageRefC<UnaryGroundLabel>> kernels;
    };

} // namespace Deconvolution

#endif // H_DECONVOLUTION_DATASET_H
