/* This file is part of the "Regression Tree Fields" (RTF) source code distribution,
 * obtained from http://www.gris.tu-darmstadt.de/research/visinf/software.
 * It is provided to you under the terms of the Microsoft Research License Agreement
 * (MSR-LA). Please see License.txt for details.
 *
 *
 * File: Deconvolution.cpp
 * Implements learning and inference for a discriminative image deconvolution application.
 *
 * ---
 * Last modified by Uwe Schmidt on 08/05/2014.
 */

namespace Deconvolution
{
    template<typename TFeature, typename TUnaryGroundLabel> class DeconvolutionOperator;
}
#define TYPE_DECONVOLUTION_OPERATOR 42
#define INSTANTIATE_CUSTOM_OPERATOR(typeid) new Deconvolution::DeconvolutionOperator<TFeature, TUnaryGroundLabel>()

#ifdef USE_MPI
#pragma message("MPI support enabled")
#endif

// #define CACHING_TYPE ON_THE_FLY
#define CACHING_TYPE WEIGHTS_AND_BASIS_PRECOMPUTED

#include "RTF/Types.h"
#include "RTF/Basic.h"
#include "RTF/Stacked.h"
#include "RTF/Monitor.h"
#include "RTF/Utility.h"
#include "RTF/Criteria.h"
#include "RTF/Learning.h"
#include "RTF/Training.h"
#include "RTF/Classify.h"
#include "RTF/Serialization.h"

#include "Dataset.h"
#include "Features.h"
#include "Operator.h"
// boundary pixels that should not be used to compute the loss, this should be set
// to (at least) half the largest kernel size in the training set
#define LOSS_BNDRY 15
#include "Loss.h"

#include <boost/filesystem.hpp>

// Console output channel
typedef Monitor::DefaultMonitor MyMonitor;

// Important training parameters
int NumFeatures              = 512;
int PairwiseDepth            = 7;
int UnaryDepth               = 7;
int PairwiseMinPoints        = 2048;
int UnaryMinPoints           = 1024;
double UnarySmallestEigen    = 1e-2;
double UnaryLargestEigen     = 1e2;
double PairwiseSmallestEigen = 1e-2;
double PairwiseLargestEigen  = 1e2;
bool JointTraining           = true;
int StackDepth               = -1;

// Other training params
const auto Interimit      = 50;
const auto Finalit        = 10;
const auto Gradeps        = 1e-6;
const auto Staged         = false;
const auto Cgit           = 1000;
const auto Cgeps          = 1e-4;
const auto Jointsubsample = 0.50; // Subsampling: use only 50% of the pixels

typedef
Stacked::RTF<Deconvolution::FeatureSampler,
        Deconvolution::Dataset,
        GradientNormCriterion,
        true,
        true,
        NullPrior,
        Monitor::DefaultMonitor,
        CACHING_TYPE> RTFtype;

// Various options for spatial layout of the pairwise factors
const Vector2D<int>
PairwiseOffsets3x3[]         = { Vector2D<int>(1,0), Vector2D<int>(0,1),
                                 Vector2D<int>(1,1), Vector2D<int>(-1,1)
                               };

const Vector2D<int>
PairwiseOffsets5x5[]         = { Vector2D<int>(-2,-2), Vector2D<int>(-1,-2), Vector2D<int>(0, -2), Vector2D<int>(1, -2), Vector2D<int>(2, -2),
                                 Vector2D<int>(-2,-1), Vector2D<int>(-1,-1), Vector2D<int>(0, -1), Vector2D<int>(1, -1), Vector2D<int>(2, -1),
                                 Vector2D<int>(-2, 0), Vector2D<int>(-1, 0)
                               };

const Vector2D<int>*
PairwiseOffsets              = PairwiseOffsets5x5;
size_t NumPairwise           = sizeof(PairwiseOffsets5x5)/sizeof(Vector2D<int>);

// Input and output paths
std::string DataPath         = "";
std::string OutputPath       = "";

std::string ModelFile        = "";
std::string InputFile        = "";
std::string OutputFile       = "";

// Returns the name of a loss
template<typename TLossTag> std::string LossName()
{
    return Loss::Loss<Traits<Deconvolution::FeatureSampler, Deconvolution::Dataset>, TLossTag>::Name();
}

// Clear existing performance statistics files before appending
void ClearStatistics(Deconvolution::Dataset& testds, const std::string& systemName, const std::string& dsName)
{
    boost::filesystem::path outp = OutputPath;

#ifdef USE_MPI
    if( MPI::Communicator().rank() == 0 )
    {
#endif
        std::ofstream overall((outp / ("Overall " + dsName + " " + systemName + ".txt")).string());
        overall.close();
#ifdef USE_MPI
    }
#endif

    for( int i = 0; i < testds.GetImageCount(); ++i )
    {
        std::ofstream individual((outp / (testds.GetImageName(i) + " " + systemName + ".txt")).string());
        individual.close();
    }
}

// Collects performance statistics given a dataset and the predictions made thereon
template<typename TLossTag, typename TModel>
void
Evaluate(Deconvolution::Dataset& testds, const VecCRef<ImageRefC<Deconvolution::Dataset::UnaryGroundLabel>>& prediction,
         TModel& model, const std::string& systemName, const std::string& dsName)
{
    boost::filesystem::path outp = OutputPath;

    // Compute overall performance statistics
    const auto overallLoss = model.template EvaluateMicroAveraged<TLossTag>(testds, prediction);
#ifdef USE_MPI
    if( MPI::Communicator().rank() == 0 )
    {
#endif
        std::ofstream overall((outp / ("Overall " + dsName + " " + systemName + ".txt")).string(), std::ios_base::app);
        overall << LossName<TLossTag>() << ": " << overallLoss << std::endl;
        overall.close();
#ifdef USE_MPI
    }
#endif

    // Compute per-image performance statistics
    for( int i = 0; i < testds.GetImageCount(); ++i )
    {
        auto loss = model.Evaluate<TLossTag>(testds.GetGroundTruthImage(i), prediction[i]);
        std::ofstream individual((outp / (testds.GetImageName(i) + " " + systemName + ".txt")).string(), std::ios_base::app);
        individual << LossName<TLossTag>() << ": " << loss << std::endl;
        testds.SaveGroundTruthImage(prediction[i], (outp / (testds.GetImageName(i) + " " + systemName + ".png")).string());
        testds.SaveGroundTruthImage(prediction[i], (outp / (testds.GetImageName(i) + " " + systemName + ".dlm")).string());
    }
}

// Reports all relevant options and collects performance statistics on test and training data
template<typename TModel>
void
EvaluateAll(TModel& model, const std::string& systemName, long long trainTicks)
{
    boost::filesystem::path outp = OutputPath;

#ifdef USE_MPI
    if( MPI::Communicator().rank() == 0 )
    {
#endif
        std::ofstream options((outp / ("Options " + systemName + ".txt")).string());
        options << "NumFeatures: " << NumFeatures << std::endl;
        options << "UnaryDepth : " << UnaryDepth << std::endl;
        options << "PairwiseDepth: " << PairwiseDepth << std::endl;
        options << "UnaryMinPoints: " << UnaryMinPoints << std::endl;
        options << "PairwiseMinPoints: " << PairwiseMinPoints << std::endl;
        options << "UnarySmallestEigen: " << UnarySmallestEigen << std::endl;
        options << "UnaryLargestEigen: " << UnaryLargestEigen << std::endl;
        options << "PairwiseSmallestEigen: " << PairwiseSmallestEigen << std::endl;
        options << "PairwiseLargestEigen: " << PairwiseLargestEigen << std::endl;
        options << "JointTraining: " << JointTraining << std::endl;
        if( PairwiseOffsets == PairwiseOffsets3x3 )
            options << "Connectivity: " << "3x3" << std::endl;
        else if( PairwiseOffsets == PairwiseOffsets5x5 )
            options << "Connectivity: " << "5x5" << std::endl;
        else
            options << "Connectivity: " << "1x1" << std::endl;
        options << "DataPath: " << DataPath << std::endl;
        options << "OutputPath: " << OutputPath << std::endl;
        options.close();
#ifdef USE_MPI
    }
#endif

    Deconvolution::Dataset testds(DataPath, "test");
    auto startTick = GetTickCountPortable();
    auto testPrediction = model.Regress(testds, Cgit, Cgeps);
    const auto timePrediction = (GetTickCountPortable() - startTick);

#ifdef USE_MPI
    if( MPI::Communicator().rank() == 0 )
    {
#endif
        std::ofstream timing((outp / ("Timing " + systemName + ".txt")).string());
        timing << "Training: " << trainTicks << " ms" << std::endl;
        timing << "Predicting for " << testds.GetImageCount() << " test images: " << timePrediction << " ms" << std::endl;
        timing.close();
#ifdef USE_MPI
    }
#endif

    ClearStatistics(testds, systemName, "Test");
    // Evaluate<Loss::PSNR>(testds, testPrediction, model, systemName, "Test");
    Evaluate<Loss::CentralPSNR>(testds, testPrediction, model, systemName, "Test");
    testds.ReleaseCache();

    Deconvolution::Dataset trainds(DataPath, "train");
    auto trainPrediction = model.Regress(trainds, Cgit, Cgeps);
    ClearStatistics(trainds, systemName, "Train");
    // Evaluate<Loss::PSNR>(trainds, trainPrediction, model, systemName, "Train");
    Evaluate<Loss::CentralPSNR>(trainds, trainPrediction, model, systemName, "Train");
    trainds.ReleaseCache();

#ifdef USE_MPI
    if( MPI::Communicator().rank() == 0 )
    {
#endif
        MyMonitor::Report("Serializing model ...\n");
        model.Serialize((outp / ("Model " + systemName + ".txt")).string());
#ifdef USE_MPI
    }
#endif
}


// Trains a cascade of RTFs in a loss-specific manner
template<typename TLossTag>
void
TrainAndEvaluate_StackedRTF(const std::string& name)
{
    RTFtype rtf;

    // add linear operator
    rtf.AddLinearOperator(TYPE_DECONVOLUTION_OPERATOR);

    rtf.AddUnaryFactorType(NumFeatures, UnaryDepth, UnaryMinPoints, UnarySmallestEigen, UnaryLargestEigen);

    for( size_t i = 0; i < NumPairwise; ++i )
        rtf.AddPairwiseFactorType(PairwiseOffsets[i], NumFeatures, PairwiseDepth, PairwiseMinPoints,
                                  PairwiseSmallestEigen, PairwiseLargestEigen);

    auto startTicks = GetTickCountPortable();
    Deconvolution::Dataset traindb(DataPath, "train");

    for( size_t n = 0; n < StackDepth; ++n )
    {
        if (n > 0)
            MyMonitor::Report("Successfully trained and evaluated model %u, now learning next model to add to the stack.\n", n);

        rtf.template LearnOneMoreDiscriminative<TLossTag, true>(traindb, Interimit, Finalit, Gradeps, Staged, Cgit, Cgeps, Jointsubsample);

        EvaluateAll(rtf, name + "-stage" + std::to_string(n+1), GetTickCountPortable()-startTicks);
    }
}

// Write ground truth and input images to output directory (this is useful for browsing the results, but not really needed)
void SaveGroundTruthAndInput(Deconvolution::Dataset& ds)
{
    boost::filesystem::path outp = OutputPath;

    for( int i = 0; i < ds.GetImageCount(); ++i )
    {
        ds.SaveGroundTruthImage(ds.GetGroundTruthImage(i), (outp / (ds.GetImageName(i) + " Ground.png")).string());
        ds.SaveInputImage(ds.GetInputImage(i), (outp / (ds.GetImageName(i) + " Input.png")).string());
    }
}

// Run the selected experiment
template<typename TOp>
void Run(TOp TrainAndEvaluate)
{
    TrainAndEvaluate();

    // Save ground truth and input along with predictions
    Deconvolution::Dataset trainds(DataPath, "train");
    SaveGroundTruthAndInput(trainds);
    trainds.ReleaseCache();

    Deconvolution::Dataset testds(DataPath, "test");
    SaveGroundTruthAndInput(testds);
    testds.ReleaseCache();
}

template<typename TModel>
void PredictAndSave(int stage, TModel& rtf, Deconvolution::Dataset& testds)
{
    boost::filesystem::path outp = OutputPath;
    auto prediction = rtf.Regress(testds, Cgit, Cgeps, stage);
    for( int i = 0; i < testds.GetImageCount(); ++i )
    {
        const auto outname = testds.GetImageName(i) + "-stage" + std::to_string(stage+1);
        testds.SavePrediction(prediction, i, (outp / (outname + ".png")).string());
        testds.SavePrediction(prediction, i, (outp / (outname + ".dlm")).string());
    }
}

// Create a prediction for the single given input file (stacked model)
void PredictStacked()
{
    boost::filesystem::path p = DataPath;
    std::cout << "ModelFile  : " << ModelFile << std::endl;
    std::cout << "DataPath   : " << DataPath << std::endl;
    std::cout << "InputFile  : " << InputFile << " (-> " << (p / (InputFile + ".txt")).string() << ")" << std::endl;
    std::cout << "OutputPath : " << OutputPath << std::endl;
    std::cout << "Stage      : " << StackDepth << std::endl;
    std::cout << std::endl;
    
    // load RTF cascade from model file   
    RTFtype rtf(ModelFile);

    // TODO: very inefficient to redo prediction of all previous stages when predicting later ones
    Deconvolution::Dataset testds(DataPath, InputFile);

    if (StackDepth == -1) { // hasn't been set by the user
        for (int stage = 0; stage < rtf.EnsembleSize(); ++stage)
            PredictAndSave(stage, rtf, testds);
    } else {
        PredictAndSave(StackDepth-1, rtf, testds);
    }

    testds.ReleaseCache();
}

// Command line parsing

void Usage(const std::string& name)
{
    std::cerr << "Usage: Deconvolution " << name << " <inputdir> <outputdir> <connectivity> <treedepth> <stackdepth>" << std::endl;
    ::exit(1);
}

void UsagePredict(const std::string& name)
{
    std::cerr << "Usage: Deconvolution " << name << " <modelfile> <inputdir> <inputfile> <outputdir> [<stackdepth>]" << std::endl;
    ::exit(1);
}

void SetConnectivity(const std::string& conn, const std::string& name)
{
    if( conn == "1x1" )
    {
        PairwiseOffsets = NULL;
        NumPairwise     = 0;
    }
    else if( conn == "3x3" )
    {
        PairwiseOffsets = PairwiseOffsets3x3;
        NumPairwise     = sizeof(PairwiseOffsets3x3)/sizeof(Vector2D<int>);
    }
    else if ( conn == "5x5" )
    {
        PairwiseOffsets = PairwiseOffsets5x5;
        NumPairwise     = sizeof(PairwiseOffsets5x5)/sizeof(Vector2D<int>);
    }
    else
    {
        MyMonitor::Report("Failed to parse <connectivity> argument (must be one of: 1x1, 3x3, 5x5).\n");
        Usage(name);
    }
}

void SetTreeDepth(const std::string& depth, const std::string& name)
{
    int max_depth;
    int success = sscanf(depth.c_str(), "%d", &max_depth);
    if (success != 1)
    {
        MyMonitor::Report("Failed to parse <treedepth> argument.\n");
        Usage(name);
    }
    else
    {
        UnaryDepth = max_depth;
        PairwiseDepth = max_depth;
    }
}

void SetStackDepth(const std::string& depth, const std::string& name)
{
    int stack_depth;
    int success = sscanf(depth.c_str(), "%d", &stack_depth);
    if (success != 1)
    {
        MyMonitor::Report("Failed to parse <stackdepth> argument.\n");
        Usage(name);
    }
    else
    {
        if( stack_depth < 1 )
        {
            MyMonitor::Report("Stack depth must be at least 1\n");
            Usage(name);
        }
        StackDepth = stack_depth;
    }
}

void SetDataPath(const std::string& rawpath, const std::string& name)
{
    DataPath = rawpath;
    if( ! boost::filesystem::exists(DataPath) )
    {
        throw Deconvolution::IOException("Data path " + DataPath + " does not exist!");
    }
}

void SetOutputPath(const std::string& rawpath, const std::string& name)
{
    OutputPath = rawpath;

    if( ! boost::filesystem::exists(OutputPath) )
    {
        throw Deconvolution::IOException("Output path " + OutputPath + " does not exist!");
    }
}

void SetModelFile(const std::string& rawpath, const std::string& name)
{
    ModelFile = rawpath;

    if( ! boost::filesystem::exists(rawpath) )
    {
        throw Deconvolution::IOException("Model file " + ModelFile + " does not exist!");
    }
}

void SetInputFile(const std::string& rawpath, const std::string& name)
{
    InputFile = rawpath;
    boost::filesystem::path p = DataPath;
    std::string file = (p / (InputFile + ".txt")).string();
    if( ! boost::filesystem::exists(file) )
    {
        throw Deconvolution::IOException("Input file " + file + " does not exist!");
    }
}

void SetOutputFile(const std::string& rawpath, const std::string& name)
{
    OutputFile = rawpath;

    if( ! boost::filesystem::exists(rawpath) )
    {
     throw Deconvolution::IOException("Output path " + rawpath + " does not exist!");
    }       
}

std::string ParseArgumentsPredictStacked(int argc, char* argv[], const std::string& name)
{
    int arg = 2;

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <modelfile> argument\n");
        UsagePredict(name);
    }
    std::string model = (argv[arg++]);
    SetModelFile(model, name);

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <inputdir> argument\n");
        UsagePredict(name);
    }
    std::string data = (argv[arg++]);
    SetDataPath(data, name);    

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <inputfile> argument\n");
        UsagePredict(name);
    }
    std::string input = (argv[arg++]);
    SetInputFile(input, name);

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <outputfile> argument\n");
        UsagePredict(name);
    }
    std::string output = (argv[arg++]);
    SetOutputPath(output, name);

    if( argc <= arg )
    {
        return name;
    }
    else
    {
        std::string depth = (argv[arg++]);
        SetStackDepth(depth, name);
    }

    return name;
}

std::string ParseArgumentsStacked(int argc, char* argv[], const std::string& name)
{
    int arg = 2;

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <inputdir> argument\n");
        Usage(name);
    }
    std::string input = (argv[arg++]);
    SetDataPath(input, name);

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <outputdir> argument\n");
        Usage(name);
    }
    std::string output = (argv[arg++]);
    SetOutputPath(output, name);

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <connectivity> argument\n");
        Usage(name);
    }
    std::string conn = (argv[arg++]);
    SetConnectivity(conn, name);

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <treedepth> argument\n");
        Usage(name);
    }
    std::string tree_depth = (argv[arg++]);
    SetTreeDepth(tree_depth, name);

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <stackdepth> argument\n");
        Usage(name);
    }
    std::string stack_depth = (argv[arg++]);
    SetStackDepth(stack_depth, name);

    return name + "_" + conn + "_" + tree_depth + "_" + stack_depth;
}

int main(int argc, char* argv[])
{
#ifdef USE_MPI
    // Initialize MPI
    MPI::Environment(argc, argv);
    auto& comm = MPI::Communicator();

    MyMonitor::Report("This is MPI process no. %d to ground control.\n", comm.rank());
#endif

    try
    {
        const char * systemStrings =    "RTFStackedForCentralPSNR, "
                                        "PredictStacked";
        if( argc < 2 )
        {
            MyMonitor::Report("Missing <system> argument (must be one of: %s)\n", systemStrings);
            ::exit(1);
        }
        std::string sys = (argv[1]);

        if ( sys == "RTFStackedForCentralPSNR" )
        {
            auto name = ParseArgumentsStacked(argc, argv, sys);
            Run([&]()
            {
                TrainAndEvaluate_StackedRTF<Loss::CentralPSNR>(name);
            });
        }
        else if ( sys == "PredictStacked" )
        {
            ParseArgumentsPredictStacked(argc, argv, sys);
            PredictStacked();
        }
        else
        {
            MyMonitor::Report("Invalid <system> argument (must be one of: %s)\n", systemStrings);
            ::exit(1);
        }

    }
    catch ( std::exception& e )
    {
        std::cerr << e.what() << std::endl;
        ::exit(1);
    }
    return 0;
}
