/* This file is part of the "Regression Tree Fields" (RTF) source code distribution,
 * obtained from http://www.gris.tu-darmstadt.de/research/visinf/software.
 * It is provided to you under the terms of the Microsoft Research License Agreement
 * (MSR-LA). Please see License.txt for details.
 *
 *
 * File: DeconvolutionCVPR13.cpp
 * Implements image deconvolution inference for learned models from the CVPR 2013 paper.
 *
 * ---
 * Last modified by Uwe Schmidt on 08/05/2014.
 */

namespace Deconvolution
{
    template<typename TFeature, typename TUnaryGroundLabel> class DeconvolutionOperator;
}
#define TYPE_DECONVOLUTION_OPERATOR 42
#define INSTANTIATE_CUSTOM_OPERATOR(typeid) new Deconvolution::DeconvolutionOperator<TFeature, TUnaryGroundLabel>()


// disable MPI
#undef USE_MPI

// #define CACHING_TYPE ON_THE_FLY
#define CACHING_TYPE WEIGHTS_AND_BASIS_PRECOMPUTED

#include "RTF/Types.h"
#include "RTF/Stacked.h"
#include "RTF/Monitor.h"
#include "RTF/Utility.h"
#include "RTF/Classify.h"

// to load learned models from CVPR'13 paper
#define REPRODUCE_CVPR13_RESULTS
#include "Dataset.h"
#include "Features.h"
#include "Operator.h"

#include <boost/filesystem.hpp>

typedef Monitor::DefaultMonitor MyMonitor;

const auto Cgit           = 1000;
const auto Cgeps          = 1e-4;

std::string DataPath         = "";
std::string OutputPath       = "";
std::string ModelFile        = "";
std::string InputFile        = "";
int StackDepth               = -1;

typedef
Stacked::RTF<Deconvolution::FeatureSampler,
        Deconvolution::Dataset,
        GradientNormCriterion,
        true,
        true,
        NullPrior,
        Monitor::DefaultMonitor,
        CACHING_TYPE> RTFtype;


template<typename TModel>
void PredictAndSave(int stage, TModel& rtf, Deconvolution::Dataset& testds)
{
    boost::filesystem::path outp = OutputPath;
    auto prediction = rtf.Regress(testds, Cgit, Cgeps, stage);
    for( int i = 0; i < testds.GetImageCount(); ++i )
    {
        const auto outname = testds.GetImageName(i) + "-rtf" + std::to_string(stage+1);
        testds.SavePrediction(prediction, i, (outp / (outname + ".png")).string());
        testds.SavePrediction(prediction, i, (outp / (outname + ".dlm")).string());
    }
}


// Create a prediction for the single given input file (stacked model)
void PredictStacked()
{
    boost::filesystem::path p = DataPath;
    std::cout << "ModelFile  : " << ModelFile << std::endl;
    std::cout << "DataPath   : " << DataPath << std::endl;
    std::cout << "InputFile  : " << InputFile << " (-> " << (p / (InputFile + ".txt")).string() << ")" << std::endl;
    std::cout << "OutputPath : " << OutputPath << std::endl;
    std::cout << "Stage      : " << StackDepth << std::endl;
    std::cout << std::endl;
 
    // load RTF cascade from model file   
    RTFtype rtf(ModelFile);

    // TODO: inefficient to redo prediction of all previous stages when predicting later ones...
    Deconvolution::Dataset testds(DataPath, InputFile);

    if (StackDepth == -1) { // hasn't been set by the user
        for (int stage = 0; stage < rtf.EnsembleSize(); ++stage)
            PredictAndSave(stage, rtf, testds);
    } else {
        PredictAndSave(StackDepth-1, rtf, testds);
    }

    testds.ReleaseCache();
}


void Usage()
{
    std::cerr << "Usage: DeconvolutionCVPR13 <ModelFile> <DataPath> <InputFile> <OutputPath> [<Stage>]" << std::endl;
}

void SetStackDepth(const std::string& depth)
{
    int stack_depth;
    int success = sscanf(depth.c_str(), "%d", &stack_depth);
    if (success != 1)
    {
        MyMonitor::Report("Failed to parse <stackdepth> argument.\n");
        Usage();
    }
    else
    {
        if( stack_depth < 1 )
        {
            MyMonitor::Report("Stack depth must be at least 1\n");
            Usage();
        }
        StackDepth = stack_depth;
    }
}

void SetDataPath(const std::string& rawpath)
{
    DataPath = rawpath;
    if( ! boost::filesystem::exists(DataPath) )
    {
        throw Deconvolution::IOException("Data path " + DataPath + " does not exist!");
    }
}

void SetOutputPath(const std::string& rawpath)
{
    OutputPath = rawpath;
    if( ! boost::filesystem::exists(OutputPath) )
    {
        throw Deconvolution::IOException("Output path " + OutputPath + " does not exist!");
    }
}

void SetModelFile(const std::string& rawpath)
{
    ModelFile = rawpath;
    if( ! boost::filesystem::exists(rawpath) )
    {
        throw Deconvolution::IOException("Model file " + ModelFile + " does not exist!");
    }
}

void SetInputFile(const std::string& rawpath)
{
    InputFile = rawpath;
    boost::filesystem::path p = DataPath;
    std::string file = (p / (InputFile + ".txt")).string();
    if( ! boost::filesystem::exists(file) )
    {
        throw Deconvolution::IOException("Input file " + file + " does not exist!");
    }
}

bool ParseArguments(int argc, char* argv[])
{
    int arg = 1;

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <ModelFile> argument\n");
        Usage();
        return false;
    }
    std::string model = (argv[arg++]);
    SetModelFile(model);

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <DataPath> argument\n");
        Usage();
        return false;
    }
    std::string data = (argv[arg++]);
    SetDataPath(data);

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <InputFile> argument\n");
        Usage();
        return false;
    }
    std::string input = (argv[arg++]);
    SetInputFile(input);

    if( argc <= arg )
    {
        MyMonitor::Report("Missing <OutputPath> argument\n");
        Usage();
        return false;
    }
    std::string output = (argv[arg++]);
    SetOutputPath(output);

    if( argc > arg )
    {
        std::string depth = (argv[arg++]);
        SetStackDepth(depth);
    }
    return true;
}

int main(int argc, char* argv[])
{
    try
    {
        if (ParseArguments(argc, argv))
            PredictStacked();
        else
            return 1;
    }
    catch ( std::exception& e )
    {
        std::cerr << e.what() << std::endl;
        return 1;
    }
    return 0;
}
