/* This file is part of the "Regression Tree Fields" (RTF) source code distribution,
 * obtained from http://www.gris.tu-darmstadt.de/research/visinf/software.
 * It is provided to you under the terms of the Microsoft Research License Agreement
 * (MSR-LA). Please see License.txt for details.
 *
 *
 * File: Loss.h
 * Defines loss functions suitable for deconvolution.
 *
 * ---
 * Last modified by Uwe Schmidt on 07/05/2014.
 */


#ifndef H_DECONVOLUTION_LOSS_H
#define H_DECONVOLUTION_LOSS_H

#include "RTF/Loss.h"

#ifndef LOSS_BNDRY
#define LOSS_BNDRY 0
#pragma message ( "LOSS_BNDRY should be defined before including Loss.h" )
#endif

namespace Loss
{
    class CentralMSE;
    class CentralPSNR;

    namespace Detail
    {

        template<typename TTraits>
        struct CentralMSE
        {
            static typename TTraits::ValueType
            Objective(const ImageRefC<typename TTraits::UnaryGroundLabel>& ground,
                      const ImageRefC<typename TTraits::UnaryGroundLabel>& prediction)
            {
                const auto cy = ground.Height(), cx = ground.Width();
                const auto cc = TTraits::UnaryGroundLabel::Size;
                typename TTraits::ValueType loss = 0;

                #pragma omp parallel for
                for(int y = LOSS_BNDRY; y < cy-LOSS_BNDRY; ++y)
                {
                    typename TTraits::ValueType lineLoss = 0;

                    for(int x = LOSS_BNDRY; x < cx-LOSS_BNDRY; ++x)
                    {
                        for(int c = 0; c < cc; ++c)
                        {
                            auto delta = ground(x, y)[c] - prediction(x, y)[c];
                            lineLoss  += delta * delta;
                        }
                    }

                    #pragma omp atomic
                    loss += lineLoss;
                }

                return loss;
            }

            static void
            Gradient(const ImageRefC<typename TTraits::UnaryGroundLabel>& ground,
                     const ImageRefC<typename TTraits::UnaryGroundLabel>& prediction,
                     const ImageRef<typename TTraits::UnaryGroundLabel>& gradient)
            {
                const auto cy = ground.Height(), cx = ground.Width();
                const auto cc = TTraits::UnaryGroundLabel::Size;

                // 0-init (might not be necessary)
                #pragma omp parallel for
                for(int y = 0; y < cy; ++y)
                    for(int x = 0; x < cx; ++x)
                        for(int c = 0; c < cc; ++c)
                            gradient(x, y)[c] = 0;

                #pragma omp parallel for
                for(int y = LOSS_BNDRY; y < cy-LOSS_BNDRY; ++y)
                {
                    for(int x = LOSS_BNDRY; x < cx-LOSS_BNDRY; ++x)
                    {
                        for(int c = 0; c < cc; ++c)
                        {
                            auto delta       = ground(x, y)[c] - prediction(x, y)[c];
                            gradient(x, y)[c] = -2.0 * delta;
                        }
                    }
                }
            }

            static typename TTraits::ValueType NormalizationConstant(const ImageRefC<typename TTraits::UnaryGroundLabel>& ground)
            {
                return (ground.Width()-2*LOSS_BNDRY) * (ground.Height()-2*LOSS_BNDRY) * TTraits::UnaryGroundLabel::Size;
            }

            static const char* Name()
            {
                return ("CentralMSE(" + std::to_string(LOSS_BNDRY) + ")").c_str();
            }

            static bool RequiresDiscreteInference()
            {
                return false;
            }
        };

        template<typename TTraits>
        struct CentralPSNR
        {
            static typename TTraits::ValueType
            Objective(const ImageRefC<typename TTraits::UnaryGroundLabel>& ground,
                      const ImageRefC<typename TTraits::UnaryGroundLabel>& prediction)
            {
                const auto loss = CentralMSE<TTraits>::Objective(ground, prediction);
                const auto normc = CentralMSE<TTraits>::NormalizationConstant(ground);
                return 10.0 * (std::log10(loss) - std::log10(normc));
            }

            static void
            Gradient(const ImageRefC<typename TTraits::UnaryGroundLabel>& ground,
                     const ImageRefC<typename TTraits::UnaryGroundLabel>& prediction,
                     const ImageRef<typename TTraits::UnaryGroundLabel>& gradient)
            {
                const auto cy = ground.Height(), cx = ground.Width();
                const auto cc = TTraits::UnaryGroundLabel::Size;
                CentralMSE<TTraits>::Gradient(ground, prediction, gradient);
                const auto loss  = CentralMSE<TTraits>::Objective(ground, prediction);
                const auto scale = 10.0 / (loss * std::log(10.0));

                #pragma omp parallel for
                for(int y = LOSS_BNDRY; y < cy-LOSS_BNDRY; ++y)
                {
                    for(int x = LOSS_BNDRY; x < cx-LOSS_BNDRY; ++x)
                    {
                        for(int c = 0; c < cc; ++c)
                        {
                            gradient(x, y)[c] *= scale;
                        }
                    }
                }
            }

            static typename TTraits::ValueType NormalizationConstant(const ImageRefC<typename TTraits::UnaryGroundLabel>& ground)
            {
                return 1;
            }

            static const char* Name()
            {
                return ("CentralPSNR(" + std::to_string(LOSS_BNDRY) + ")").c_str();
            }

            static bool RequiresDiscreteInference()
            {
                return false;
            }
        };        

        // Template specializations that choose the proper loss class based on a loss tag
        template<typename TTraits>
        struct LossDispatcher<TTraits, ::Loss::CentralMSE>
        {
            typedef ::Loss::Detail::CentralMSE<TTraits> Loss;
        };

        template<typename TTraits>
        struct LossDispatcher<TTraits, ::Loss::CentralPSNR>
        {
            typedef ::Loss::Detail::CentralPSNR<TTraits> Loss;
        };


    }

}

#endif // H_DECONVOLUTION_LOSS_H