/* This file is part of the "Regression Tree Fields" (RTF) source code distribution,
 * obtained from http://www.gris.tu-darmstadt.de/research/visinf/software.
 * It is provided to you under the terms of the Microsoft Research License Agreement
 * (MSR-LA). Please see License.txt for details.
 *
 *
 * File: Operator.h
 * Implements a custom linear operator for deconvolution.
 *
 * ---
 * Last modified by Uwe Schmidt on 07/05/2014.
 */


#ifndef H_DECONVOLUTION_OPERATOR_H
#define H_DECONVOLUTION_OPERATOR_H

#include <itkImage.h>
#include <itkFFTConvolutionImageFilter.h>
#include <itkConstantBoundaryCondition.h>
#include <itkFlipImageFilter.h>
#include <itkMaskImageFilter.h>

#include "RTF/Utility.h"
#include "RTF/Compute.h"
#include "RTF/LinearOperator.h"

namespace Deconvolution
{
    template <typename TFeature, typename TUnaryGroundLabel>
    class DeconvolutionOperator : public LinearOperator::OperatorBase<TFeature, TUnaryGroundLabel, LinearOperator::DefaultWeights<typename TUnaryGroundLabel::ValueType>>
    {
    public:
        static const size_t VarDim   = TUnaryGroundLabel::Size;
        static const size_t BasisDim = TFeature::UnaryBasisSize;
        typedef typename TUnaryGroundLabel::ValueType TValue;

        typedef Compute::SystemVectorRef<TValue, VarDim>  SystemVectorRef;
        typedef Compute::SystemVectorCRef<TValue, VarDim> SystemVectorCRef;
        typedef typename SystemVectorRef::VectorType      VectorType;

    private:
        typedef itk::Image<double, 2> ImageType;
        typedef itk::FFTConvolutionImageFilter<ImageType> ConvolutionFilterType;
        typedef itk::ConstantBoundaryCondition<ImageType, ImageType> BoundaryConditionType;      
        typedef itk::FlipImageFilter<ImageType> FlipFilterType;
        typedef itk::MaskImageFilter<ImageType, ImageType> MaskFilterType;

        const ImageType::Pointer _image, _kernel, _mask;
        const ConvolutionFilterType::Pointer conv2_filter_rhs, conv2_filter1, conv2_filter2;
        const FlipFilterType::Pointer flip_filter;
        const MaskFilterType::Pointer mask_filter;
        BoundaryConditionType conv2_bc;
        typename FlipFilterType::FlipAxesArrayType axesArray;
        // intial value important for learning
        double _alpha;

        inline const size_t BlurKernelSize(const typename TFeature::PreProcessType& prep, int pos) const
        {
            // size of blur kernel in last two pixels
            const auto cx = prep.Width(), cy = prep.Height();
            const auto kernel_idx = BasisDim-1;
            return static_cast<size_t>(prep(cx-1-pos,cy-1)[kernel_idx]);
        }

        void CreateImageAndMask(const ImageType::Pointer image, const ImageType::Pointer mask, const typename TFeature::PreProcessType& prep, const SystemVectorCRef& data) const
        {
            // image size
            const auto cx = prep.Width(), cy = prep.Height();
            // kernel size, to mask pixels at boundary
            const auto ckx = BlurKernelSize(prep,1), cky = BlurKernelSize(prep,0);
            const auto bx = (ckx-1)/2, by = (cky-1)/2;
            // Copy data
            #pragma omp parallel for
            for(int y = 0; y < cy; y++)
            {
                for(int x = 0; x < cx; x++)
                {
                    const ImageType::IndexType idx = {{x,y}};
                    const bool mask_cond = x<bx || y<by || cx-x-1<bx || cy-y-1<by;
                    image->SetPixel(idx, data(x,y)[0]);
                    mask->SetPixel(idx, mask_cond ? 0.0 : 1.0);
                }
            }
        }

        void CreateMaskedImage(const ImageType::Pointer image, const typename TFeature::PreProcessType& prep) const
        {
            // image size
            const auto cx = prep.Width(), cy = prep.Height();
            // kernel size, to mask pixels at boundary
            const auto ckx = BlurKernelSize(prep,1), cky = BlurKernelSize(prep,0);
            const auto bx = (ckx-1)/2, by = (cky-1)/2;
            // Copy data
            #pragma omp parallel for
            for(int y = 0; y < cy; y++)
            {
                for(int x = 0; x < cx; x++)
                {
                    const ImageType::IndexType idx = {{x,y}};
                    const bool mask_cond = x<bx || y<by || cx-x-1<bx || cy-y-1<by;
                    image->SetPixel(idx, mask_cond ? 0.0 : prep(x,y)[0]);
                }
            }
        }      

        void CreateKernelAndPrepareImageAndMask(const ImageType::Pointer image, const ImageType::Pointer kernel, const ImageType::Pointer mask, const typename TFeature::PreProcessType& prep) const
        {
            // reset images
            image->Initialize(); kernel->Initialize(); mask->Initialize();
            // image size
            const auto cx = prep.Width(), cy = prep.Height();
            // kernel size, to mask pixels at boundary
            const auto ckx = BlurKernelSize(prep,1), cky = BlurKernelSize(prep,0);
            const auto bx = (ckx-1)/2, by = (cky-1)/2;

            // kernel setup
            ImageType::IndexType start = {{0,0}};
            ImageType::SizeType size;
            size[0] = ckx;
            size[1] = cky;

            ImageType::RegionType region;
            region.SetSize(size);
            region.SetIndex(start);

            kernel->SetRegions(region);
            kernel->Allocate();
            
            // Copy data
            #pragma omp parallel for
            for(int y = 0; y < cky; y++)
            {
                for(int x = 0; x < ckx; x++)
                {
                    ImageType::IndexType idx = {{x,y}};
                    kernel->SetPixel(idx, prep(x,y)[BasisDim-1]);
                }
            }        

            // image and mask setup
            size[0] = cx;
            size[1] = cy;
            region.SetSize(size);

            image->SetRegions(region);
            image->Allocate();

            mask->SetRegions(region);
            mask->Allocate();
        }      

        void MultiplyAndAdd(const SystemVectorRef& res, const ImageType::Pointer out) const
        {
            // important: call Update() to actually trigger execution of the ITK pipeline
            out->Update();

            const auto image_size = out->GetLargestPossibleRegion().GetSize();
            const auto cx = image_size[0], cy = image_size[1];

            #pragma omp parallel for
            for (int y = 0; y < cy; y++)
            {
                for (int x = 0; x < cx; x++)
                {
                    const ImageType::IndexType idx = {{x,y}};
                    res(x,y)[0] += _alpha * out->GetPixel(idx);
                }
            }
        }      

    public:
        DeconvolutionOperator()
        :   conv2_filter_rhs(ConvolutionFilterType::New()), conv2_filter1(ConvolutionFilterType::New()), conv2_filter2(ConvolutionFilterType::New()),
            conv2_bc(BoundaryConditionType()), flip_filter(FlipFilterType::New()), mask_filter(MaskFilterType::New()),
            _image(ImageType::New()), _kernel(ImageType::New()), _mask(ImageType::New()), _alpha(1000)
        {
            conv2_bc.SetConstant(0);
            conv2_filter_rhs->SetBoundaryCondition(&conv2_bc);
            conv2_filter1->SetBoundaryCondition(&conv2_bc);
            conv2_filter2->SetBoundaryCondition(&conv2_bc);
            // note: ITK 'convolution' actually computes correlation
            axesArray.Fill(true);
            flip_filter->SetFlipAxes(axesArray);

            // ITK pipeline setup
            flip_filter->SetInput(_kernel);
            // rhs
            conv2_filter_rhs->SetInput(_image);
            conv2_filter_rhs->SetKernelImage(flip_filter->GetOutput());
            // matrix
            mask_filter->SetMaskImage(_mask);
            conv2_filter1->SetInput(_image);
            conv2_filter1->SetKernelImage(_kernel);
            mask_filter->SetInput(conv2_filter1->GetOutput());
            conv2_filter2->SetInput(mask_filter->GetOutput());
            conv2_filter2->SetKernelImage(flip_filter->GetOutput());
        }

        void AddInImplicitMatrixMultipliedBy(const typename TFeature::PreProcessType& prep,
                const SystemVectorRef& Fy, const SystemVectorCRef& y) const
        {
            // multiplication with equation system matrix
            CreateKernelAndPrepareImageAndMask(_image, _kernel, _mask, prep);
            CreateImageAndMask(_image, _mask, prep, y);
            MultiplyAndAdd(Fy,conv2_filter2->GetOutput());
        }

        void AddInLinearContribution(const typename TFeature::PreProcessType& prep, const SystemVectorRef& l) const
        {
            // for RHS of equation system
            CreateKernelAndPrepareImageAndMask(_image, _kernel, _mask, prep);
            CreateMaskedImage(_image, prep);
            MultiplyAndAdd(l,conv2_filter_rhs->GetOutput());
        }

        const TValue* SetNonDifferentiableParameters(const TValue* alpha)
        {
            _alpha = *alpha;
            return alpha;
        }        

        void Print() const {};

        int Type() const
        {
            return TYPE_DECONVOLUTION_OPERATOR;
        }

        std::istream& Deserialize(std::istream& in)
        {
            in >> _alpha;
            return in;
        }

        std::ostream& Serialize(std::ostream& out) const
        {
            out << _alpha << std::endl;
            return out;
        }

        // Learning of _alpha currently not supported
        
        const TValue* CheckFeasibility(const TValue *ws, bool& feasible) const { return ws; }

        TValue* Project(TValue *ws) const { return ws; }

        size_t NumWeights() const { return 0; }

        const TValue* SetWeights(const TValue *ws) { return ws; }

        TValue* GetWeights(TValue *ws) const { return ws; }

        TValue* GetGradientAddPrior(TValue *gs, TValue& objective) const { return gs; }

        void ClearGradient() {}

        void ResetWeights() {}

#ifdef _OPENMP
        void InitializeLocks() {};
        void DestroyLocks() {};
#endif      

    };

}

#endif // H_DECONVOLUTION_OPERATOR_H
